@@ -1,4 +1,5 @@
 import asyncio
+import datetime
 
 from asgiref.sync import sync_to_async
 from django.db import models as django_models
@@ -14,11 +15,11 @@ def default_cache_func(
         filter_by,
 ):
     redis_dicts = redis_root.get(redis_model, return_dict=True)
-    # cache_to_django(
-    #     django_model,
-    #     redis_dicts,
-    #     redis_root,
-    # )
+    cache_to_django(
+        django_model,
+        redis_dicts,
+        redis_root,
+    )
     django_to_cache(
         django_model,
         redis_model,
@@ -48,7 +49,8 @@ def cache_to_django(django_model, redis_dicts, redis_root):
             if len(async_tasks) == async_chunk_size or completed == to_complete:
                 await asyncio.gather(*async_tasks)
                 async_tasks = []
-                print(f'Recorded {completed} {django_model.__name__} instances from cache to django')
+                print(
+                    f'{datetime.datetime.now()} - Written {completed} {django_model.__name__} instances from cache to django')
 
     loop = asyncio.get_event_loop()
     loop.run_until_complete(run_cache_to_django(django_model, redis_dicts, redis_root))
@@ -59,71 +61,126 @@ async def update_or_create_django_instance_from_redis_instance(
         django_model,
 ):
     django_id = redis_instance['django_id']
-    django_params, django_many_to_many_params = redis_dict_to_django_params(
+    django_params, django_many_to_many_params = await async_redis_dict_to_django_params(
         redis_instance,
-        django_model
+        django_model,
     )
     new_django_params = {}
     for k, v in django_params.items():
-        if k != 'django_id':
+        if k not in ['django_id', 'id']:
             new_django_params[k] = v
     django_params = new_django_params
-    django_instance = None
-    if django_id != 'null':
-        if django_model.objects.filter(id=django_id):
-            django_instance = django_model.objects.filter(id=django_id).update(
-                **django_params
-            )
-            update_django_many_to_many(
-                django_model,
-                django_id,
-                django_many_to_many_params
-            )
-    if django_instance is None:
-        django_instance = django_model.objects.create(
-            **django_params
-        )
-        update_django_many_to_many(
-            django_model,
-            django_id,
-            django_many_to_many_params
-        )
+    django_instance = await sync_to_async_update_or_create_django_instance_from_redis_instance(
+        django_id,
+        django_model,
+        django_params,
+        django_many_to_many_params,
+    )
+
     return django_instance
 
 
-def redis_dict_to_django_params(
+@sync_to_async
+def sync_to_async_update_or_create_django_instance_from_redis_instance(
+        django_id,
+        django_model,
+        django_params,
+        django_many_to_many_params,
+):
+    need_to_create = True
+    if django_id != 'null':
+        if django_model.objects.filter(id=django_id):
+            need_to_create = False
+
+    print(need_to_create, django_params, django_many_to_many_params)
+
+    django_instance = None
+
+    if need_to_create:
+        pass
+    #     django_instance = django_model.objects.create(**django_params)
+    #     update_django_many_to_many(
+    #         django_model,
+    #         django_instance.id,
+    #         django_many_to_many_params
+    #     )
+    else:
+        django_instance = django_model.objects.filter(id=django_id)[0]
+        changed_fields_to_update = {}
+        for redis_field_name, redis_field_value in django_params.items():
+            django_instance_fields_value = getattr(django_instance, redis_field_name)
+            if django_instance_fields_value != redis_field_value:
+                changed_fields_to_update[redis_field_name] = redis_field_value
+        changed_many_to_many_fields_to_update = {}
+        for redis_field_name, redis_field_value in django_many_to_many_params.items():
+            django_instance_fields_value = getattr(django_instance, redis_field_name)
+            if django_instance_fields_value != redis_field_value:
+                changed_many_to_many_fields_to_update[redis_field_name] = redis_field_value
+        print(changed_fields_to_update, changed_many_to_many_fields_to_update)
+    #     django_model.objects.filter(id=django_id).update(**changed_fields_to_update)
+    #     update_django_many_to_many(
+    #         django_model,
+    #         django_id,
+    #         changed_many_to_many_fields_to_update
+    #     )
+    #     django_instance = django_model.objects.filter(id=django_id)[0]
+    return django_instance
+
+
+async def async_redis_dict_to_django_params(
         redis_dict,
         django_model
 ):
     django_params = {}
     django_many_to_many_params = {}
-    for django_field in django_model._meta.get_fields():
-        django_field_name = django_field.name
+    django_model_fields = await sync_to_async_get_model_fields(django_model)
+    for django_field in django_model_fields:
+        django_field_data = await sync_to_async_get_django_field_data(django_field)
+        django_field_name = django_field_data['name']
         if django_field_name in redis_dict.keys():
-            if django_field.__class__ == django_models.ForeignKey:
-                foreign_key_model = django_field.remote_field.model
+            if django_field_data['class'] == django_models.ForeignKey:
+                foreign_key_model = django_field_data['remote_field_model']
                 redis_foreign_key_instance = redis_dict[django_field_name]
-                django_foreign_key_instance = update_or_create_django_instance_from_redis_instance(
+                django_foreign_key_instance = await update_or_create_django_instance_from_redis_instance(
                     redis_foreign_key_instance,
                     foreign_key_model
                 )
                 django_params[django_field_name] = django_foreign_key_instance
-            elif django_field.__class__ == django_models.ManyToManyField:
-                many_to_many_model = django_field.remote_field.model
+            elif django_field_data['class'] == django_models.ManyToManyField:
+                many_to_many_model = django_field_data['remote_field_model']
                 django_many_to_many_instances = []
-                redis_many_to_many_instances = redis_dict[django_field.name]
+                redis_many_to_many_instances = redis_dict[django_field_data['name']]
                 for redis_many_to_many_instance in redis_many_to_many_instances:
-                    django_many_to_many_instance = update_or_create_django_instance_from_redis_instance(
+                    django_many_to_many_instance = await update_or_create_django_instance_from_redis_instance(
                         redis_many_to_many_instance,
                         many_to_many_model
                     )
                     django_many_to_many_instances.append(django_many_to_many_instance)
                 django_many_to_many_params[django_field_name] = django_many_to_many_instances
             else:
-                value = redis_dict[django_field.name]
+                value = redis_dict[django_field_data['name']]
                 django_params[django_field_name] = value
+
     return django_params, django_many_to_many_params
 
+@sync_to_async
+def sync_to_async_get_model_fields(django_model):
+    return list(django_model._meta.get_fields())
+
+
+@sync_to_async
+def sync_to_async_get_django_field_data(django_field):
+    django_field_data = {
+        'name': django_field.name,
+        'class': django_field.__class__,
+        'remote_field_model': None,
+    }
+    try:
+        django_field_data['remote_field_model'] = django_field.remote_field.model
+    except:
+        pass
+    return django_field_data
+
 
 def update_django_many_to_many(
         django_model,
@@ -144,12 +201,33 @@ def django_sync_to_async_list(
         django_func,
         params_to_use=None,
 ):
-    if params_to_use is None:
-        result_qs = django_func()
+    if type(params_to_use) == dict:
+        objects_qs = django_func(**params_to_use)
     else:
-        result_qs = django_func(**params_to_use)
+        objects_qs = django_func()
 
-    result = list(result_qs)
+    result = list(objects_qs)
+
+    return result
+
+
+@sync_to_async
+def django_sync_to_async_func_and_update(
+        django_func,
+        params_to_use=None,
+        params_to_update=None,
+):
+    if type(params_to_use) == dict:
+        objects_qs = django_func(**params_to_use)
+    else:
+        objects_qs = django_func()
+
+    if type(params_to_update) == dict:
+        result = objects_qs.update(**params_to_update)
+    else:
+        result = objects_qs
+
+    result = list(result)
 
     return result
 
@@ -163,6 +241,16 @@ def django_sync_to_async_get(
     return param
 
 
+@sync_to_async
+def django_sync_to_async_create(
+        django_model,
+        params
+):
+    new_django_instance = django_model.objects.create(**params)
+    new_django_instance_id = new_django_instance.id
+    return new_django_instance_id
+
+
 def django_to_cache(
         django_model,
         redis_model,
@@ -208,7 +296,8 @@ def django_to_cache(
             if len(async_tasks) == async_chunk_size or completed == to_complete:
                 await asyncio.gather(*async_tasks)
                 async_tasks = []
-                print(f'Recorded {completed} {django_model.__name__} instances from django to cache')
+                print(
+                    f'{datetime.datetime.now()} - Written {completed} {django_model.__name__} instances from django to cache')
 
     loop = asyncio.get_event_loop()
     loop.run_until_complete(
@@ -248,9 +337,7 @@ async def update_or_create_redis_instance_from_django_instance(
             django_id=django_instance.id,
             **fields_to_create
         ).save()
-        print(redis_model.__name__, redis_instance['django_id'])
         result_redis_instance_id = redis_instance['id']
-        print(f'{redis_model} with django_id={django_instance.id} created...')
     else:
         redis_instance = redis_instance[0]
         fields_to_update = check_fields_need_to_update(
@@ -261,7 +348,6 @@ async def update_or_create_redis_instance_from_django_instance(
                           **fields_to_update
                           )
         result_redis_instance_id = redis_instance['id']
-        print(f'{redis_model} with django_id={django_instance.id} existing...')
 
     return result_redis_instance_id
 
@@ -341,17 +427,25 @@ async def django_instance_to_redis_params(
         exclude_fields,
 ):
     redis_params = {}
+    django_model = django_instance.__class__
     for i, django_field in enumerate(django_instance.__class__._meta.get_fields()):
         if not django_field.__class__.__name__.endswith('Rel'):
-            redis_param = await django_instance_field_to_redis_value(
-                django_instance,
-                django_field,
-                redis_root,
-                save_related_models,
-                get_or_create_redis_model_from_django_model,
-                exclude_fields,
-            )
-            redis_params[django_field.name] = redis_param
+            django_field_name = django_field.name
+            allowed = True
+            if django_model in exclude_fields.keys():
+                allowed = (django_field_name not in exclude_fields[django_model])
+            if allowed:
+                if django_field.name not in ['id', 'pk']:
+                    redis_param = await django_instance_field_to_redis_value(
+                        django_instance,
+                        django_field,
+                        redis_root,
+                        save_related_models,
+                        get_or_create_redis_model_from_django_model,
+                        exclude_fields,
+                    )
+                    if redis_param:
+                        redis_params[django_field_name] = redis_param
     return redis_params
 
 
