@@ -1,3 +1,4 @@
+import asyncio
 import datetime
 import decimal
 import json
@@ -12,15 +13,12 @@ from time import sleep
 
 class RedisField:
 
-    def __init__(self, default=None, choices=None, null=True, ttl=None):
+    def __init__(self, default=None, choices=None, null=True):
         self.default = default
         self.value = None
         check_types(choices, dict)
         self.choices = choices
         self.null = null
-        if ttl is not None:
-            check_types(ttl, (int, float))
-        self.ttl = ttl
 
 
     def _get_default_value(self):
@@ -56,7 +54,7 @@ class RedisField:
         if value == 'null':
             if not self.null:
                 if redis_root.ignore_deserialization_errors:
-                    print(f'{value} can not be deserialized like {self.__class__.__name__}, ignoring')
+                    print(f'{datetime.datetime.now()} - {value} can not be deserialized like {self.__class__.__name__}, ignoring')
                 else:
                     raise Exception(f'{value} can not be deserialized like {self.__class__.__name__}')
 
@@ -91,10 +89,13 @@ class RedisNumber(RedisField):
     def deserialize_value(self, value, redis_root):
         self.deserialize_value_check_null(value, redis_root)
         if value != 'null':
-            if '.' in value:
-                value = float(value)
+            if type(value) == str:
+                if '.' in value:
+                    value = float(value)
+                else:
+                    value = int(value)
             else:
-                value = int(value)
+                check_types(value, (int, float))
         else:
             value = 0
         return value
@@ -299,8 +300,8 @@ class RedisBool(RedisNumber):
     def deserialize_value(self, value, redis_root):
         self.deserialize_value_check_null(value, redis_root)
         if value != 'null':
-            check_types(value, str)
-            value = int(value)
+            check_types(value, int)
+            value = bool(value)
         return value
 
 
@@ -309,14 +310,13 @@ class RedisDecimal(RedisNumber):
     def clean(self):
         self.value = self.check_value()
         if self.value not in [None, 'null']:
-            check_types(self.value, decimal.Decimal)
-            self.value = int(self.value)
+            check_types(self.value, (int, float, decimal.Decimal))
+            self.value = float(self.value)
         return super().clean()
 
     def deserialize_value(self, value, redis_root):
         self.deserialize_value_check_null(value, redis_root)
         if value != 'null':
-            check_types(value, str)
             value = decimal.Decimal(value)
         return value
 
@@ -417,7 +417,7 @@ class RedisRoot:
         if type(prefix) == str:
             self.prefix = prefix
         else:
-            print(f'Prefix {prefix} is type of {type(prefix)}, allowed only str, using default prefix "redis_test"')
+            print(f'{datetime.datetime.now()} - Prefix {prefix} is type of {type(prefix)}, allowed only str, using default prefix "redis_test"')
             self.prefix = 'redis_test'
         self.connection_pool = self._get_connection_pool(connection_pool)
         self.async_db_requests_limit = async_db_requests_limit
@@ -448,6 +448,8 @@ class RedisRoot:
                     'to_cache_in': now + datetime.timedelta(seconds=ttl),
                     'cache_conf': cache_conf
                 }
+        if django_models_to_cache:
+            print(f"{datetime.datetime.now()} - {', '.join([str(model) for model in django_models_to_cache.keys()])} will be cached now...")
         return django_models_to_cache
 
     def check_cache(self):
@@ -483,7 +485,7 @@ class RedisRoot:
                     if django_model in exclude_fields.keys():
                         allowed = (django_field_name not in exclude_fields[django_model])
                     if allowed:
-                        if django_field_name in ['id', 'pk']:
+                        if django_field.name in ['id', 'pk']:
                             django_field_name = 'django_id'
                             redis_field = RedisNumber(null=False)
                         else:
@@ -593,10 +595,10 @@ class RedisRoot:
                 exclude_fields,
                 filter_by,
             )
-            print(f'Cached {django_model}')
+            print(f'{datetime.datetime.now()} - Cached {django_model} successfully')
 
     def fast_get_keys_values(self, string):
-        keys = self.redis_instance.scan_iter(string)
+        keys = list(self.redis_instance.scan_iter(string))
         values = self.redis_instance.mget(keys)
         results = dict(zip(keys, values))
         return results
@@ -606,7 +608,7 @@ class RedisRoot:
             connection_pool.connection_kwargs['decode_responses'] = True
             self.connection_pool = connection_pool
         else:
-            print(f'{self.__class__.__name__}: No connection_pool provided, trying default config...')
+            print(f'{datetime.datetime.now()} - {self.__class__.__name__}: No connection_pool provided, trying default config...')
             default_host = 'localhost'
             default_port = 6379
             default_db = 0
@@ -714,7 +716,7 @@ class RedisRoot:
                 break
         if not found:
             if self.ignore_deserialization_errors:
-                print(f'{model_name} not found in registered models, ignoring')
+                print(f'{datetime.datetime.now()} - {model_name} not found in registered models, ignoring')
                 model = model_name
             else:
                 raise Exception(f'{model_name} not found in registered models')
@@ -729,7 +731,7 @@ class RedisRoot:
                 field_instance = getattr(model, field_name)
         except BaseException as ex:
             if self.ignore_deserialization_errors:
-                print(f'{model.__name__} has no field {field_name}, ignoring deserialization')
+                print(f'{datetime.datetime.now()} - {model.__name__} has no field {field_name}, ignoring deserialization')
                 field_instance = field_name
             else:
                 raise Exception(f'{model.__name__} has no field {field_name}')
@@ -741,7 +743,7 @@ class RedisRoot:
             pass
         except BaseException as ex:
             if self.ignore_deserialization_errors:
-                print(f'{raw_value} can not be deserialized like {field_instance.__class__.__name__}, ignoring')
+                print(f'{datetime.datetime.now()} - {raw_value} can not be deserialized like {field_instance.__class__.__name__}, ignoring')
                 value = raw_value
             else:
                 raise Exception(f'{raw_value} can not be deserialized like {field_instance.__class__.__name__}')
@@ -856,17 +858,19 @@ class RedisRoot:
         model_name = model.__name__
         instances_with_allowed = {}
         fast_get_results = self.fast_get_keys_values(f'{self.prefix}:{model_name}:*')
-        for key, raw_value in fast_get_results:
-            prefix, model_name, instance_id, field_name = key.split(':')
+        for instance_key, fields_json in fast_get_results.items():
+            prefix, model_name, instance_id = instance_key.split(':')
             instance_id = int(instance_id)
-            value = self.deserialize_value(raw_value, model_name, field_name)
-            allowed = self._filter_field_name(field_name, value, filters)
-            if instance_id not in instances_with_allowed.keys():
-                instances_with_allowed[instance_id] = {}
-            instances_with_allowed[instance_id][field_name] = {
-                'value': value,
-                'allowed': allowed
-            }
+            fields_dict = json.loads(fields_json)
+            for field_name, raw_value in fields_dict.items():
+                value = self.deserialize_value(raw_value, model_name, field_name)
+                allowed = self._filter_field_name(field_name, value, filters)
+                if instance_id not in instances_with_allowed.keys():
+                    instances_with_allowed[instance_id] = {}
+                instances_with_allowed[instance_id][field_name] = {
+                    'value': value,
+                    'allowed': allowed
+                }
         return instances_with_allowed
 
     def _get_instances_from_instances_with_allowed(self, instances_with_allowed):
@@ -916,21 +920,34 @@ class RedisRoot:
 
     def _update_by_key(self, key, fields_to_update, renew_ttl, new_ttl):
         updated_data = None
-        prefix, model_name, instance_id, field_name = key.split(':')
+        prefix, model_name, instance_id = key.split(':')
+        saved_model = self._get_registered_model_by_name(model_name)
         instance_id = int(instance_id)
-        if field_name in fields_to_update.keys():
-            saved_model = self._get_registered_model_by_name(model_name)
-            if issubclass(saved_model, RedisModel):
-                saved_field_instance = self._get_field_instance_by_name(field_name, saved_model)
-                if issubclass(saved_field_instance.__class__, RedisField):
-                    saved_field_instance.value = fields_to_update[field_name]
-                    cleaned_value = saved_field_instance.clean()
-                    field_ttl = self._get_ttl_by_update_params(saved_field_instance, saved_model, renew_ttl, new_ttl)
-                    self.redis_instance.set(key, cleaned_value, ex=field_ttl)
-                    updated_data = {
-                        'model': saved_model,
-                        'id': instance_id
-                    }
+        instance_data_json = self.redis_instance.get(key)
+        instance_data = json.loads(instance_data_json)
+        data_to_write = {}
+        for field_name, field_data in instance_data.items():
+            saved_field_instance = self._get_field_instance_by_name(field_name, saved_model)
+            if field_name in fields_to_update.keys():
+                saved_field_instance.value = fields_to_update[field_name]
+                cleaned_value = saved_field_instance.clean()
+            else:
+                cleaned_value = field_data
+            if key not in data_to_write.keys():
+                data_to_write[key] = {}
+            data_to_write[key][field_name] = cleaned_value
+        for instance_key, fields_to_write in data_to_write.items():
+            ttl = None
+            if renew_ttl:
+                ttl = saved_model.get_model_ttl()
+            elif new_ttl:
+                ttl = new_ttl
+            fields_to_write_json = json.dumps(fields_to_write)
+            self.redis_instance.set(instance_key, fields_to_write_json, ex=ttl)
+            updated_data = {
+                'model': saved_model,
+                'id': instance_id
+            }
         return updated_data
 
     def update(self, model, instances=None, return_dict=False, renew_ttl=False, new_ttl=None, **fields_to_update):
@@ -944,7 +961,7 @@ class RedisRoot:
         else:
             ids_to_update = get_ids_from_untyped_data(instances)
             for instance_id in ids_to_update:
-                for key in self.redis_instance.scan_iter(f'{self.prefix}:{model_name}:{instance_id}:*'):
+                for key in self.redis_instance.scan_iter(f'{self.prefix}:{model_name}:{instance_id}'):
                     updated_data = self._update_by_key(key, fields_to_update, renew_ttl, new_ttl)
                     if updated_data is not None:
                         updated_datas.append(updated_data)
@@ -971,7 +988,7 @@ class RedisRoot:
         else:
             ids_to_delete = get_ids_from_untyped_data(instances)
             for instance_id in ids_to_delete:
-                for key in self.redis_instance.scan_iter(f'{self.prefix}:{model_name}:{instance_id}:*'):
+                for key in self.redis_instance.scan_iter(f'{self.prefix}:{model_name}:{instance_id}'):
                     self._delete_by_key(key)
 
     # def get_id_from_django_id(self, model, django_id):
@@ -1061,43 +1078,40 @@ class RedisModel:
             else:
                 raise Exception(f'{self.__class__.__name__} has no field {name}')
 
-    def get_field_ttl(self, field):
-        field_ttl = field.ttl
-        if field_ttl is None:
-            meta = self.__model_data__['meta']
-            if 'ttl' in meta.keys():
-                field_ttl = meta['ttl']
-        return field_ttl
-
-    def _serialize_data(self, full=True):
+    def _serialize_data(self):
         redis_root = self.__model_data__['redis_root']
         name = self.__model_data__['name']
         fields = self.__model_data__['fields']
         fields = dict(fields)
-        model_key = f'{redis_root.prefix}:{name}:{self.id.value}'
+        instance_key = f'{redis_root.prefix}:{name}:{self.id.value}'
         cleaned_fields = {}
         for field_name, field in fields.items():
             if not field_name.startswith('__'):
                 try:
                     cleaned_value = field.clean()
-                    if full:
-                        field_ttl = self.get_field_ttl(field)
-                        cleaned_fields[model_key + f':{field_name}'] = {
-                            'value': cleaned_value,
-                            'ttl': field_ttl
-                        }
-                    else:
-                        cleaned_fields[field_name] = cleaned_value
+                    if instance_key not in cleaned_fields.keys():
+                        cleaned_fields[instance_key] = {}
+                    cleaned_fields[instance_key][field_name] = cleaned_value
                 except BaseException as ex:
                     raise Exception(f'{ex} ({name} -> {field_name})')
         fields_with_model_key = cleaned_fields
         return fields_with_model_key
+    
+    def get_model_ttl(self):
+        ttl = None
+        meta = self.__model_data__['meta']
+        if 'ttl' in meta.keys():
+            ttl = meta['ttl']
+        return ttl
 
     def _set_fields(self):
-        field_to_write = self._serialize_data()
-        redis_instance = self.__model_data__['redis_root'].redis_instance
-        for key, field_data in field_to_write.items():
-            redis_instance.set(key, field_data['value'], ex=field_data['ttl'])
+        data_to_write = self._serialize_data()
+        model_ttl = self.get_model_ttl()
+        redis_root = self.__model_data__['redis_root']
+        for instance_key, fields_dict in data_to_write.items():
+            redis_instance = redis_root.redis_instance
+            fields_json = json.dumps(fields_dict)
+            redis_instance.set(instance_key, fields_json, ex=model_ttl)
 
     def _get_new_id(self):
         redis_root = self.__model_data__['redis_root']
